/**
 * @file   allpath.c
 * @author David Bremner
 * @date   Thu Apr 18 12:22:14 2002
 * 
 * @brief generate all inclusion minimal paths of a given length in
 * the pivot complex.
 * 
 * These are the paths that are at least geodesics in the subgraph of
 * the pivot graph induced by their vertices.
 * This file is now just a driver.
 */


#include <unistd.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif
#include "error_message.h"
#include "oms_types.h"
#include "basis_equation.h"
#include "statearray.h"
#include "path.h"
#include "combinat.h"
#include "scan.h"
#include "boundary.h"

static int expandfacets=0;
static int verbose=0;
static int dim=-1,steps=-1;
static int elements=-1;
static int *pi=NULL;
static int *occured=NULL;
static be_block_p conv_eq=NULL;


static void usage(char *name){
  printf("usage: %s [-o outfile ] -d dim -k steps -n elements\n",name);
  exit(1);
}

void output(int *path, int len){
  int i;
  if (expandfacets)
    for (i=0; i< binomial(elements,dim+1); i++){
      occured[i]=0;
    }

  for (i=1; i< len;i++){
    int bindex=path[i];
    if (bindex<0)
      break;
    err_assert(bindex>0);

    if (verbose){
      int j;

      which_kset(elements,dim,bindex,pi);
      printf("[");
      for (j=0; j< dim; j++){
	printf(" %d",pi[j]);
      }
      printf("]");
    } else if (expandfacets){
      int j,b;
      for (j=0; j< be_cols(conv_eq); j++){
	b=be_bindex(conv_eq,path[i],j);
	if (!occured[b]){
	  occured[b]=1;
	  printf(" %d",b);
	}
      }
    } else {
      printf("%d\t", bindex );
    }
  }
    printf("\n");


}

#define MAX_STEPS 200

int main(int argc, char **argv)
{

  int c;
  int i;

  int source=-1,sink=-1;
  int facets;
  be_block_p pivots;
  sarray_p found;
  int *path;
  int tok;
  int step=0;
  char *initstr=0;



  while((c=getopt(argc,argv,"o:d:k:s:t:n:I:ve"))!=-1){
    switch(c){
    case 'e':
      expandfacets=1;
      break;

    case 's':
      source=atoi(optarg);
      break;
    case 't':
      sink=atoi(optarg);
      break;
    case 'd':
      dim=atoi(optarg);
      break;
    case 'k':
      steps=atoi(optarg);
      break;
    case 'n':
      elements=atoi(optarg);
      break;
    case 'I':
      initstr=strdup(optarg);
      break;
    case 'o':
      freopen(optarg,"w",stdout);
      err_assert(stdout);
      break;
    case 'v':
      verbose=1;
      break;
    default:
      usage(argv[0]);
      exit(1);
    }
  }
  

  
  facets=binomial(elements,dim);

  if (sink==-1)
    sink=facets-1;

  if (elements==-1 || dim==-1)
    err_fatal(HERE,"you must specify dimension and number of elements");

  if (steps==-1){
    steps=elements-dim;
  }

  path=calloc(steps+1,sizeof(int));
  for(i=0; i<=steps; i++)
    path[i]=-1;

  found=sa_new(facets,facets);
  for(i=0; i<sa_size(found); i++)
    sa_init(found,i,0);

  if (initstr){
    int i=0;
    if (source!=-1)
      err_fatal(HERE,"cannot specify both source and initial path");

    scan_scan_string(initstr);
    while((tok=scanlex())>0){
      if (tok != SCAN_INT){
	usage(argv[0]);
	exit(1);
      }
      sa_set(found,scanint,1);
      path[i++]=scanint;
      source=scanint;
    }
    step=i-1;

  } else {
    step=0;
    path[0]=source;
  }
  pivots=be_alloc(facets,(elements-dim)*dim,
		  facets);
  be_neighbour_init(pivots,elements,dim+1);


  conv_eq=be_alloc(facets,(elements-dim), binomial(elements,dim+1));
  be_boundary_init(conv_eq,elements,dim+1,0,facets,-1);

  if (verbose)
    pi=calloc(steps,sizeof(int));

  if (expandfacets)
    occured=calloc(binomial(elements,dim+1),sizeof(int));



  path_bt(path,step,steps,sink, pivots,found,output);
  return 0;
}


    
