/**
 * @file   variable_choice.c
 * @author David Bremner <bremner@unb.ca>
 * @date   Thu Jul  4 15:05:56 2002
 * 
 * @brief  heuristics for chirotope backtracking variable choice
 * 
 * 
 */

#include "basis_equation.h"
#include "combinat.h"
#include "debug.h"
#include "globaldef.h"
#include "int_stack.h"
#include "nalloc.h"
#include "nodedef.h"
#include "nu_pluecker.h"
#include "nu_sign_set.h"
#include "nu_signed_index.h"
#include "oms_opt.h"
#include "oms_types.h"
#include "profile.h"
#include "ridge_graph.h"
#include "statearray.h"
#include "uni.h"

static int nvc_score(na_node_p global, na_node_p state, array_index_t basis);
static void push_all_signs(params_p params, int index, int_stack_p stack);

/** 
 * Choose the first unset variable in some static order
 * 
 * 
 * @return unset index variable, or -1 if none are found
 */
static int nvc_static(sarray_p chirotope,
		  array_index_t *reorder_table ){
  int i;

  for (i=0; i<sa_size(chirotope); i++){
    int j;
    j=reorder_table[i];
    if (!sign_determined(sa_read(chirotope,j)))
      return j;
  }
  return -1;
}


static array_index_t nvc_greedy(na_node_p global, na_node_p local, int order){
  int i, best_var=-1, best_score=-1;
  sarray_p chirotope=na_access(local,SLOT_chirotope);
  for (i=0; i < sa_size(chirotope); i++){
    if (!sign_determined(sa_read(chirotope,i))) {
      int score=nvc_score(global,local,i);
      err_assert(score>0);
      if (score > best_score){
	best_score=score;
	best_var=i;
      }
    }
  }

  return best_var;

}
  

/**
   \brief Give variable a score as a branching variable.

   \param basis Which variable to evaluate
*/

static int nvc_score(na_node_p global, na_node_p state, array_index_t basis){

  be_block_p PE;		/* %%Pl\"ucker%% equations */
  sa_array_p pe_state;
  sarray_p chirotope;
  nu_sign_set_p S;		/* variables set, but not yet processed */
  params_p params;
  bool success;
  int score;


  chirotope=na_access(state,SLOT_chirotope);
  pe_state=na_access(state,SLOT_pe_state);
  PE=na_access(global,SLOT_pe);
  params=na_access(global,SLOT_params);

  switch (params->variable_order){
  case 'r':
    return 1+uni()*sa_size(chirotope);
    break;
  case 'g':
    S=nss_new(sa_size(chirotope));
    
    sa_save(chirotope);
    sa_save(pe_state);

    sa_set(chirotope,basis,SIGN_POS);

    success=npl_traverse(basis,PE,S,chirotope,pe_state);
    if (!success){
      /* failure is good */
      score=sa_size(chirotope);
    } else {
      score=nss_count(S);
    }
    nss_destroy(S);
    sa_restore(pe_state);
    sa_restore(chirotope);

    err_debug(DB_BACKTRACK,(HERE,"score=%d",score));
    return score;
    break;
  default:
    err_fatal(HERE,"unknown variable order %c\n",params->variable_order);
  }

  /*NOTREACHED*/
  return 0;
}
/** 
 * Decide which chirotope variable to branch on next, 
 * and what order to explore signs 
 * @param global global state
 * @param S search node
 * 
 * @return true, if we found a variable to branch on.
 * false otherwise. The main output is in the backtracking stack in 'state'
 *
 * 
 */
bool nvc_push_choices(na_node_p global,na_node_p local, int_stack_p bt_stack){
  
  sarray_p chirotope;
  params_p params;
  bool found=FALSE;
  array_index_t *reorder_table;
  nu_sign_set_p choice_Q=na_access(local,SLOT_choice_Q);
  array_index_t bindex;

  if (OPT_use_profile)
    prof_event(PROF_VAR_CHOICE);

  params=na_access(global,SLOT_params);
  chirotope=na_access(local,SLOT_chirotope);

  while (!found && !nss_empty(choice_Q)){
    array_index_t k=nss_deq_index(choice_Q);
    
    if (!sign_determined(sa_read(chirotope,k))){
      err_debug(DB_BACKTRACK,(HERE,"choosing %d from Q",k));
      bindex=k;
      found=TRUE;
    }
  }
  if (!found){
    if (OPT_use_profile)
      prof_event(PROF_FREE_CHOICE);
    
    switch(params->variable_order){
    case 'l':
    case 'L':
    case 'i':
      reorder_table=na_access(global,SLOT_reorder_table);
      bindex=nvc_static(chirotope, reorder_table );
      break;
    case 'r':
    case 'g':
      bindex=nvc_greedy(global,local,params->variable_order);
      break;
    default: 
      err_fatal(HERE,"unimplimented variable order %c\n",params->variable_order);
    }
    if (bindex>=0)
      err_debug(DB_BACKTRACK,(HERE,"Free choice of basis %d",bindex));

  }
  if (bindex>=0){
    push_all_signs(params,bindex,bt_stack);
    return TRUE;
  } else {
    err_debug(DB_BACKTRACK,(HERE,"No unset sign found"));
    return FALSE;
  }
}

void nvc_dump_stack(int_stack_p stack){
  int i;
  for (i=0; i< stack->top; i++){
    uint32_t j=stack->elem[i];
    
    if (j==NSI_MARKER)
      fprintf(stderr,"-mark-\n");
    else {
      int bindex=nsi_index(j);
      nu_sign_t sign= nsi_sign(j);
      fprintf(stderr,"%d %s\n",bindex,sign_name[sign]);
    }
  }
}

static void push_all_signs(params_p params, int index, int_stack_p stack){
  
  istack_push(stack,NSI_MARKER);

  switch(params->sign_order){
  case 0:
    if (!params->uniform)
      istack_push(stack,nsi_encode(index,SIGN_ZERO));

    istack_push(stack,nsi_encode(index,SIGN_NEG));
    istack_push(stack,nsi_encode(index,SIGN_POS));
    break;
  case 1:
    istack_push(stack,nsi_encode(index,SIGN_NEG));
    istack_push(stack,nsi_encode(index,SIGN_POS));
    if (!params->uniform)
      istack_push(stack,nsi_encode(index,SIGN_ZERO));
    break;
  case 2:
    if (!params->uniform)
      istack_push(stack,nsi_encode(index,SIGN_ZERO));
    istack_push(stack,nsi_encode(index,SIGN_POS));

    istack_push(stack,nsi_encode(index,SIGN_NEG));
    break;

  default:
    err_fatal(HERE,"unknown sign order %d",params->sign_order);
  }
}
	      
	      

