/* est_main.c

   Euclidean spanning trees: main program

   Ambros Marzetta

 * $Source$
 * $Revision: 2732 $
 * $Date: 2000-11-15 17:14:52 +0100 (Wed, 15 Nov 2000) $
 * $Log$
 * Revision 1.2  2000/11/15 16:14:51  bremner
 * committing patched version of zram source
 *
 * Revision 1.2  1996/09/17  13:34:14  marzetta
 * changed type of getopt result to int
 *
 * Revision 1.1  1996/06/03  18:59:52  marzetta
 * Initial revision
 *
 *
 */

#include <stdlib.h>		/* exit() */
#include <stdio.h>
#include <sys/types.h>
#include <unistd.h>
#include <fcntl.h>
#include "z_base.h"
#include "z_rs.h"
#include "z_enumps.h"
#include "z_par.h"		/* z_myproc */
#include "z_time.h"
#include "z_ps2d.h"

#define est_pointT	unsigned short int /* number of points */
#define est_coordT	float	/* coordinate in Euclidean space */
				/* must be compatible to scanf format */
#define MAXDEGREE	100	/* used for static arrays */

int est_NodeTag;

typedef struct est_Node {
  Z_NODE_HEADER;		/* len; tag = est_NodeTag */
  est_pointT est_data[1];	/* variable size */
				/* est_right[]: right point of edge */
				/* est_left[]: 	left point of edge */
				/* always a spanning tree with n-1 edges */
				/* the edges are sorted lexicographically */
} est_Node, *est_NodeP;

#define est_right(global, p)	(p->est_data)
#define est_left(global, p)	(p->est_data + global->numPt)

#define is_last_node(x)	(!x)

typedef z_ps2dT est_Global, *est_GlobalP;

#define est_coord_x(gp, i)	ps2d_coord_x(gp,i)
#define est_coord_y(gp, i)	ps2d_coord_y(gp,i)
				/* must be adjacent in memory for qsort() */

static z_rs_ProcSet est_procset1, est_procset2, est_procset3;
static int est_procsetnum1, est_procsetnum2, est_procsetnum3;

est_NodeP est_NewNode(int numPt)
{
  est_NodeP n;

  n = (est_NodeP) z_NewNode(est_NodeTag,
			    sizeof(est_Node) + 2 * numPt * sizeof(est_pointT));
  return n;
} /* est_NewNode */

void est_FreeNode(est_NodeP node)
{
  z_FreeNode((z_NodeP) node);
} /* est_FreeNode */

boolean est_Equal(est_GlobalP global,
		  est_NodeP x,
		  est_NodeP y)
{
  int i;

  /* This assumes that the edges are sorted. */
  assert(global && global->tag == z_ps2dTag);
  assert(x);
  assert(y);
  for (i=0; i<global->numPt-1; i++) {
    if (est_right(global, x)[i] != est_right(global, y)[i]) return 0;
    if (est_left(global, x)[i] != est_left(global, y)[i]) return 0;
  }
  return 1;
} /* est_Equal */

void est_Print(est_GlobalP global, est_NodeP node)
{
  int i;

  assert(global && global->tag == z_ps2dTag);
  printf("[%2d][", z_myproc);
  if (node)
    for (i=0; i<global->numPt-1; i++)
      printf("(%2d,%2d)",
	     est_right(global, node)[i], est_left(global, node)[i]);
  else
    printf("NULL");
  printf("]\n");
} /* est_Print */

typedef struct {est_coordT x, y; } coordPair;

static est_coordT x0, y0;

static int angle_compare(const coordPair *a, const coordPair *b)
{
  float ta, tb, diff;

  /* handle points with x coordinate == x0 */
  if (a->x == x0) diff =  a->y - y0;
  else if (b->x == x0) diff =  y0 - b->y;
  else {
    ta = (a->y - y0) / (a->x - x0);
    tb = (b->y - y0) / (b->x - x0);
    /*printf("%f %f %f %f\n", a->x, a->y, ta, tb);*/
    diff = ta - tb;
  }
  if (diff > 0) return 1;
  else if (diff == 0) return 0;
  else return -1;
} /* angle_compare */

#define MAXHEADERLINE 128

est_GlobalP est_Read(FILE *infile)
     /* Read Point set from infile and allocate an est_GlobalP */
     /* Should test for 3 collinear points */
{
  char s[MAXHEADERLINE], c;
  int points, matches, line, i, i0;
  est_GlobalP global;

  points = 0;
  line = 0;			/* line numbers start with 1 */
  do {				/* read header line */
    (void) fgets(s, MAXHEADERLINE, infile);
    line ++;
    matches = sscanf(s, " Points:%d", &points);
    if (matches == 0) {
      if (sscanf(s, " %c", &c) == 1 && c != '#') {
	fprintf(stderr, "est_Read: Error in line %1d\n", line);
	return NULL;
      }
    } else if (feof(infile)) {
      fprintf(stderr, "est_Read: End of file before line %d\n", line);
      return NULL;
    }
  } while (points==0);
  
  global = z_ps2d_New(points);
  
  for (;;) {
    if (fscanf(infile, "%1s", &c) != 1)
      break;
    switch(c) {
    case 'C':
      z_ps2d_Read(infile, global);
      break;
    default:
      fprintf(stderr, "est_Read: Illegal character '%c'\n", c);
      break;
    } /* switch(ch) */
  } /* for */
  { /* find leftmost point */
    i0 = 0;
    x0 = est_coord_x(global, i0);
    y0 = est_coord_y(global, i0);
    for (i = 1; i < points; i++)
      if (est_coord_x(global, i) < x0 ||
	  (est_coord_x(global, i) == x0 && est_coord_y(global, i) < y0)) {
	i0 = i;
	x0 = est_coord_x(global, i);
	y0 = est_coord_y(global, i);
      }
  }
  { /* move leftmost point to index 0 */
    est_coord_x(global, i0) = est_coord_x(global, 0);
    est_coord_y(global, i0) = est_coord_y(global, 0);
    est_coord_x(global, 0) = x0;
    est_coord_y(global, 0) = y0;
  }
  qsort(&est_coord_x(global, 1),
	points - 1,
	sizeof(est_coordT) * 2,
	(int (*)(const void *, const void *))angle_compare);
  
  return global;
} /* est_Read */

#define BILLION 1000000000
#define INSIDE_FRACTION (1/12.0) /* distance from outermost point to border */

static void bound(est_coordT x, est_coordT *min, est_coordT *max)
{
  /* Enlarge the interval [min..max] such that x is in the interval */
  if (x < *min) *min = x;
  if (x > *max) *max = x;
} /* bound */

void est_PSHeader(est_GlobalP global, FILE *psfile, float xsize, float ysize)
{
  est_coordT xmin, xmax, ymin, ymax;
  int i;
  float ixsize, iysize;
  
  assert(global && global->tag == z_ps2dTag);
  xmin = BILLION;
  xmax = -BILLION;
  ymin = BILLION;
  ymax = -BILLION;
  for (i=0; i<global->numPt; i++) {
    bound(est_coord_x(global,i), &xmin, &xmax);
    bound(est_coord_y(global,i), &ymin, &ymax);
  }
  printf("%%bounds: [%f %f] [%f %f]\n", xmin, xmax, ymin, ymax);
  if (xmax == xmin) xmax ++;	/* to avoid division by zero */
  if (ymax == ymin) ymax ++;
  ixsize = xmax - xmin;		/* final coordinate system */
  iysize = ymax - ymin;

  fprintf(psfile, "/radius %f def\n",
	  1 / xsize * ixsize);
  fprintf(psfile, "/points {\n");
  fprintf(psfile, "  %f setlinewidth\n",
	  0.3 * ixsize / xsize * (1 + 2 * INSIDE_FRACTION));
  fprintf(psfile, "  %f %f scale\n",
	  1 / ixsize * xsize / (1 + 2 * INSIDE_FRACTION),
	  1 / iysize * ysize / (1 + 2 * INSIDE_FRACTION));
  fprintf(psfile, "  %f %f translate\n",
	  ixsize * INSIDE_FRACTION - xmin , iysize * INSIDE_FRACTION - ymin);
  for (i=0; i<global->numPt; i++) {
    fprintf(psfile, "  newpath %f %f radius 0 360 arc fill\n",
	    est_coord_x(global,i),
	    est_coord_y(global,i));
  }
  fprintf(psfile, "} def\n");
} /* est_PSHeader */

void est_PSPicture(est_GlobalP global, est_NodeP node, FILE *psfile,
		   float xsize, float ysize)
{
  int i;

  assert(global && global->tag == z_ps2dTag);
  assert(node && node->tag == est_NodeTag);
  fprintf(psfile, "points\n");	/* defined in est_PSHeader */
  for (i=0; i<global->numPt-1; i++) { /* ith edge of tree */
    fprintf(psfile, "newpath %f %f moveto %f %f lineto stroke\n",
	    est_coord_x(global, est_right(global, node)[i]),
	    est_coord_y(global, est_right(global, node)[i]),
	    est_coord_x(global, est_left(global, node)[i]),
	    est_coord_y(global, est_left(global, node)[i]));
  } /* for i */
} /* est_PSPicture */

static float right(est_GlobalP global, int a, int b, int p)
{
  /* return a positive number if p is to the right of (a,b) */
  return (est_coord_y(global, b) - est_coord_y(global, a)) *
    (est_coord_x(global, p) - est_coord_x(global, a)) -
    (est_coord_x(global, b) - est_coord_x(global, a)) *
    (est_coord_y(global, p) - est_coord_y(global, a));
} /* right */

static boolean est_Dominates(est_GlobalP global, est_NodeP node,
			     est_pointT i, est_pointT j)
{
  est_pointT p, q, r, s;	/* point numbers */
  /* Does edge i dominate edge j? */
  assert(global);
  assert(node);
  assert(i > 0);
  assert(j > 0);
  assert((long)i < (long)global->numPt - 1);
  assert((long)j < (long)global->numPt - 1);
  assert(i != j);
  p = est_right(global, node)[i];
  q = est_left(global, node)[i];
  r = est_right(global, node)[j];
  s = est_left(global, node)[j];
  if (p >= s || r >= q) return 0; /* intervals do not intersect */
  /* p < s && r < q */
  if (p <= r && s <= q)		/* (p,q) contains (r,s) */
    if (p != r)
      return right(global, p,q,r) > 0;
    else
      return right(global, p,q,s) > 0;
  if (p < r) 			/* r < q < s: test leftdom */
    return right(global, r,s,q) < 0;
  /* r <= p < s: test rightdom*/
    return right(global, r,s,p) < 0;
} /* est_Dominates */

#define MAX 100
static est_pointT a[MAX][MAX], f[MAX];

static boolean DFS(int root, int from, int pt)
{
  /* is root connected to pt? */
  int next;
  if (root == pt) return 1;
  while (f[root]) {
    -- f[root];
    next = a[root][f[root]];
    if (next != from && DFS(next, root, pt)) return 1;
  }
  return 0;	/* no other edges from root */
} /* DFS */

static boolean Connected0(est_GlobalP global, est_NodeP node,
			  est_pointT without, est_pointT pt)
{
  /* Is point pt connected with point 0? Edge without is not used for
     the path. */
  int i;
  est_pointT r, l;

  assert(global->numPt < MAX);
  for (i = 0; i < global->numPt; i++)
    f[i] = 0;
  for (i = 0; i < global->numPt - 1; i++)
    if (i != without) {
      r = est_right(global, node)[i];
      l = est_left(global, node)[i];
      a[r][f[r]++] = l;
      a[l][f[l]++] = r;
    }
  return DFS(0, -1, pt);
} /* Connected0 */

#define edge_greater(a,b,e) \
  (a > est_right(global, node)[e] || \
   (a == est_right(global, node)[e] && b > est_left(global, node)[e]))
  
static void Replace(est_GlobalP global, est_NodeP node,
		    est_pointT e, est_pointT a, est_pointT b)
{
  int i;
  /* Remove edge number e. Insert edge (a,b). Sort edges. */
  assert(a < b);
  if (e == 0 || edge_greater(a,b,e-1)) {
    /* Insert at position e or later */
    i = e+1;
    while (i < global->numPt-1   && edge_greater(a,b,i)) {
      est_right(global, node)[i-1] = est_right(global, node)[i];
      est_left(global, node)[i-1] = est_left(global, node)[i];
      i++;
    } ;
    /* Insert at position i-1 */
    est_right(global, node)[i-1] = a;
    est_left(global, node)[i-1] = b;
  } else {			/* Insert at position 0..e-1 */
    i = e-1;
    do {
      est_right(global, node)[i+1] = est_right(global, node)[i];
      est_left(global, node)[i+1] = est_left(global, node)[i];
      --i;
    } while (i >= 0 && ! edge_greater(a,b,i));
    /* Insert at position i+1 */
    est_right(global, node)[i+1] = a;
    est_left(global, node)[i+1] = b;
  }
} /* Replace */

static est_NodeP est_f(est_GlobalP global, est_NodeP node)
{
  /* Allocate f(node).
     f(node) = exchange lexicographically smallest edge which is not dominated.
   */
  est_pointT best, i;
  est_NodeP f_node;

  assert(global && global->tag == z_ps2dTag);
  assert(node && node->tag == est_NodeTag);
  best = 0;
  while ((long)best < (long)global->numPt - 1 &&
	 est_right(global, node)[best] == 0)
    best ++;

  assert(est_right(global, node)[best] != 0); /* must not be root */

  i = best + 1;
  while ((long)i < (long)global->numPt - 1 &&
	 est_right(global, node)[i] < est_left(global, node)[best]) {
    if (est_Dominates(global, node, i, best)) best = i;
    i ++;
  }

  /*fprintf(stderr, "best = %d\n", best);*/
  /* remove edge number best */
  f_node = (est_NodeP) z_NewCopyNode((z_NodeP) node);
  if (Connected0(global, node, best, est_left(global, node)[best]))
				/* replace by (0, right) */
    Replace(global, f_node, best, 0, est_right(global, node)[best]);
  else				/* replace by (0, left) */
    Replace(global, f_node, best, 0, est_left(global, node)[best]);
  
  return f_node;
} /* est_f */

static boolean is_root_node(est_GlobalP global, est_NodeP node)
{
  /* edges are sorted */
  assert(global && global->tag == z_ps2dTag);
  assert(node && node->tag == est_NodeTag);
  return est_right(global, node)[global->numPt - 2] == 0;
} /* is_root_node */

static boolean Intersects(est_GlobalP global, est_NodeP node,
		    est_pointT i, est_pointT a, est_pointT b)
{
  /* Does edge i intersect (a,b)?
     Precondition: i != (a,b) */
  est_pointT p,q;

  p = est_right(global, node)[i];
  q = est_left(global, node)[i];

  if (p == a || p == b || q == a || q == b) return 0;
  return right(global, a, b, p) * right(global, a, b, q) < 0 &&
         right(global, p, q, a) * right(global, p, q, b) < 0;
} /* Intersects */

static est_NodeP est_adj_oracle(est_GlobalP global,
				est_NodeP node,
				unsigned int k)
{
  /* Allocate neighbour k of node, if it exists. This oracle is not
     symmetric, but maybe more efficient than a symmetric one. 
     Return NULL if neighbour k does not exist.
     k must be in [0..(n-1)*(n-2)] */
  
  est_NodeP oracle = NULL;
  int i;
  est_pointT old, new, pivot, a, b;

  assert(global && global->tag == z_ps2dTag);
  assert(node && node->tag == est_NodeTag);
  if (k == 0)
    if (is_root_node(global, node))	
      return NULL;
    else
      return est_f(global, node);
  old = (k-1) / (global->numPt - 2);
  assert((long)old < (long)global->numPt - 1);
  if (est_right(global, node)[old] != 0) return NULL;
  pivot = est_left(global, node)[old]; /* common point of old and new edge */
  new = (k-1) % (global->numPt - 2) + 1; /* other point of new edge */
  if (new >= pivot) {
    b = new+1;
    a = pivot;
  } else {
    b = pivot;
    a = new;
  }
  /* a < b */
  /* Replace edge (0,pivot) by (a,b) */

  for (i=global->numPt-2; i>=0; --i)
    if (Intersects(global, node, i, a, b)) return NULL;
  oracle = (est_NodeP) z_NewCopyNode((z_NodeP) node);
  Replace(global, oracle, old, a, b);
  if (! Connected0(global, oracle, -1, pivot)) {
    est_FreeNode(oracle);
    return NULL;
  }
  
  return oracle;
} /* est_adj_oracle */

static est_NodeP est_Root(est_GlobalP global)
     /* Allocate and initialize the root of the reverse search tree */
{
  est_NodeP root;
  int i;			/* must be signed */

  root = est_NewNode(global->numPt);
  for (i = global->numPt - 2; i >= 0; --i) {
    est_right(global, root)[i] = 0;
    est_left(global, root)[i] = i+1;
  }
  return root;
} /* est_Root */

unsigned int est_Delta(est_GlobalP global, est_NodeP this)
{
  return (global->numPt-1)*(global->numPt-2) + 1;
} /* est_Delta */

void initproc(void)
{
  /* This function is called once on every processing element. */

  z_enumps_init();
  z_rs_Init();
  z_ps2d_Init();

  est_NodeTag = z_InstallClass(est_Node, 0, 0, 0);
  
  est_procset1.rs_Equal = (z_EqualProc)est_Equal;
  est_procset1.rs_Condition = Z_CONDITION_ALWAYS;
  est_procset1.rs_Result = (z_NodeProc)est_Print;
  est_procset1.rs_F = (z_rs_FProc)est_f;
  est_procset1.rs_AdjOracle = (z_rs_AdjOracleProc)est_adj_oracle;
  est_procset1.rs_IsRoot = (z_ConditionProc)is_root_node;
  est_procset1.rs_Delta = (z_rs_DeltaProc)est_Delta;
  est_procsetnum1 = z_InstallProcSet((z_ProcSetP) &est_procset1);

  est_procset2.rs_Equal = (z_EqualProc)est_Equal;
  est_procset2.rs_Condition = Z_CONDITION_ALWAYS;
  est_procset2.rs_Result = z_Save;
  est_procset2.rs_F = (z_rs_FProc)est_f;
  est_procset2.rs_AdjOracle = (z_rs_AdjOracleProc)est_adj_oracle;
  est_procset2.rs_IsRoot = (z_ConditionProc)is_root_node;
  est_procset2.rs_Delta = (z_rs_DeltaProc)est_Delta;
  est_procsetnum2 = z_InstallProcSet((z_ProcSetP) &est_procset2);

  est_procset3.rs_Equal = (z_EqualProc)est_Equal;
  est_procset3.rs_Condition = Z_CONDITION_ALWAYS;
  est_procset3.rs_Result = NULL;
  est_procset3.rs_F = (z_rs_FProc)est_f;
  est_procset3.rs_AdjOracle = (z_rs_AdjOracleProc)est_adj_oracle;
  est_procset3.rs_IsRoot = (z_ConditionProc)is_root_node;
  est_procset3.rs_Delta = (z_rs_DeltaProc)est_Delta;
  est_procsetnum3 = z_InstallProcSet((z_ProcSetP) &est_procset3);

} /* initproc */

static int Usage(const char *name)
{
  fprintf(stderr, "Usage: %s [-1npt]\n\n", name);
  fprintf(stderr, "Options: -1  sequential\n");
  fprintf(stderr, "         -h  help (produce this message)\n");
  fprintf(stderr, "         -n  no output (implies time measurement)\n");
  fprintf(stderr, "         -p  PostScript output\n");
  fprintf(stderr, "         -t  time measurement\n");
  return 4;
} /* Usage */

int mmain(int argc, char **argv, char **envp)
{
  est_GlobalP global;
  est_NodeP root;
  double time;
  enum {ot_normal, postScript, noOutput} outputType = ot_normal;
  boolean showTime=0;
  boolean sequential=0;
  int procsetnum;
  int c;

  while ((c=getopt(argc, argv, "1hnpt")) != -1)
    switch(c) {
    case 'p':
      outputType = postScript;
      sequential = 1;
      break;
    case 't':
      showTime = 1;
      break;
    case 'n':
      outputType = noOutput;
      showTime = 1;
      break;
    case '1':
      sequential = 1;
      break;
    default:
      return Usage(argv[0]);
    }
  
  if (argc != optind) return Usage(argv[0]);
  
  global = est_Read(stdin);
  if (!global) exit(1);
  root = est_Root(global);

  switch (outputType) {
  case noOutput:
    procsetnum = est_procsetnum3; /* no output */
    break;
  case postScript:
    procsetnum = est_procsetnum2; /* save */
    z_enumps_Open("/tmp/pm");
    break;
  default:
    procsetnum = est_procsetnum1; /* print */
    break;
  }
  
  time = z_Gtime();
  z_rs_Enumerate(procsetnum, (z_lb_GlobalP) global,
		 (z_NodeP)root, sequential);
  time = z_Gtime() - time;

  if (outputType == postScript) {
    z_PrintPS((z_NodeP)global, stdout,
	      (z_enumps_HeaderProc)est_PSHeader, (z_enumps_PictureProc)est_PSPicture);
  }
  if (showTime)
    if (sequential)
      printf("Time:%7.2f (sequential)\n",
	     time);
    else
      printf("Time:%7.2f (%2d processor%s\n",
	     time, z_numprocs, z_numprocs>1?"s)":") ");
  return 0;
} /* mmain */
