/* z_bb3.c (ZRAM: Branch and Bound in parallel, depth first, explicit stack)
   Ambros Marzetta

 * $Source$
 * $Revision: 2732 $
 * $Date: 2000-11-15 17:14:52 +0100 (Wed, 15 Nov 2000) $
 * $Log$
 * Revision 1.2  2000/11/15 16:14:52  bremner
 * committing patched version of zram source
 *
 * Revision 1.1  1996/06/03  19:01:15  marzetta
 * Initial revision
 *
 */

#define TRACENEWUB0

#include <stdio.h>
#include <stdlib.h>
#include "z_base.h"
#include "z_bb.h"
#include "z_lb.h"
#include "s_seq.h"
#include "z_par.h"
#include "z_mp.h"

#define ShortNodeString(x) glob->psp->bb_ShortNodeString ? \
  glob->psp->bb_ShortNodeString(x) : ""
  
static boolean bb3_inited;
static z_lb_ProcSet bb3_procset;
static int bb3_procsetnum;
#ifndef SEQUENTIAL
static z_JobNum initJob, getResultJob;
static z_tagT newUBTag;
#endif /* SEQUENTIAL */

/* Types --------------------------------------------------------------- */

static int bb3_GlobalTag;

typedef struct bb3_Global {
  Z_NODE_HEADER;		/* len; tag = bb3_GlobalTag */
  z_NodeP bb3_app_global;	/* application global node */
  int bb3_psn;			/* application procedure set (z_bb_ProcSetP) */
} bb3_Global, *bb3_GlobalP;

static bb3_GlobalP NewBB3Global(z_ProcSetNum procsetnum,
				z_bb_GlobalDataP ap_global)
{
  bb3_GlobalP bb3_global;

  assert(ap_global);
  bb3_global = (bb3_GlobalP) z_NewNode(bb3_GlobalTag, sizeof(bb3_Global));
  bb3_global->bb3_psn = procsetnum;
  bb3_global->bb3_app_global = ap_global;
  return bb3_global;
} /* NewBB3Global */

typedef z_NodeP bb3_WorkUnitP;

/* Global variables ----------------------------------------------------- */

typedef struct {		/* global variables of depth-first search */
  z_bb_ProcSetP psp;		/* set in workandsplit */
  bb3_GlobalP bb3_glob;		/* set in workandsplit */
  s_seqP workList;		/* set in workandsplit */
  z_NodeP solNode;		/* installed as checkpoint */
  boolean ub_changed;		/* cleared in init */
} DFEglobT;

static DFEglobT glob;

/* Procedures ----------------------------------------------------------- */

#ifndef SEQUENTIAL
static void NewUBHandler(void)
     /* message handler */
{
  z_NodeP newub;

#ifdef TRACENEWUB
    printf("[%2d]<New Upper Bound\r\n", z_myproc);
#endif
    newub = z_GetNodeBuffer();
    if (glob.solNode == NULL ||
	glob.psp->bb_CompareProc(newub, glob.solNode) < 0) {
      /* received node is better than local upper bound */
      if (glob.solNode)
	z_FreeNode(glob.solNode);
      glob.solNode = newub;
    } else
      z_FreeNode(newub);
} /* NewUBHandler */

static void BroadcastUB(void)
{
  /* Broadcast the upper bound.
     Precondition: z_TestSend
     One node should always fit into the output buffer.
     */
  /*printf("[%2d]Broadcasting upper bound\n", z_myproc);*/
  z_AddTagBuffer(newUBTag);
  z_AddNodeBuffer(glob.solNode);
  z_Ibroadcast("upper bound");
  glob.ub_changed = 0;
} /* BroadcastUB */
#endif /* SEQUENTIAL */

static void DoDFChild(z_NodeP child, DFEglobT *glob)
{
  /* This procedure is called for all children of the node just
     extracted from the sequence. The child either goes to
     glob->solNode, to the end of the sequence, or is deallocated.
     */
  assert(child);
  if (glob->solNode == NULL ||
      glob->psp->bb_CompareProc(child, glob->solNode) < 0) {
				/* better than upper bound */
    if (glob->psp->bb_IsSolution(child)) { /* solution -> new upper bound*/
      if (glob->solNode) z_FreeNode(glob->solNode);
      glob->solNode = child;
#ifndef SEQUENTIAL
      if (!glob->ub_changed && z_numprocs > 1) {
	z_QueueSender(BroadcastUB);
	glob->ub_changed = 1;
      }
#endif /* SEQUENTIAL */
      printf("[%2d]New upper bound: (%s)\n", z_myproc, ShortNodeString(child));
      fflush(stdout);
    } else
      s_seq_InsertLast(glob->workList, child); /* no solution -> stack */
    /*printf("Insert (%s)\n", ShortNodeString(child));*/
  }
  else
    z_FreeNode(child); /* not better than upper bound -> Free */
} /* DoDFChild */

static void DoDFChildInPlace(z_NodeP child, DFEglobT *glob)
{
  /* This procedure is called for all children of the node just
     extracted from the sequence. The child either goes to
     glob->solNode, to the end of the sequence, or is deallocated.
     */
  assert(child);
  if (glob->solNode == NULL ||
      glob->psp->bb_CompareProc(child, glob->solNode) < 0) {
				/* better than upper bound */
    if (glob->psp->bb_IsSolution(child)) { /* solution -> new upper bound*/
      if (glob->solNode) z_FreeNode(glob->solNode);
      glob->solNode = z_NewCopyNode(child);
#ifndef SEQUENTIAL
      if (!glob->ub_changed && z_numprocs > 1) {
	z_QueueSender(BroadcastUB);
	glob->ub_changed = 1;
      }
#endif /* SEQUENTIAL */
      printf("[%2d]New upper bound..: (%s)\n", z_myproc,
	     ShortNodeString(child));
      fflush(stdout);
    } else
      s_seq_InsertLast(glob->workList, z_NewCopyNode(child));
				/* no solution -> queue */
    /*printf("Insert (%s)\n", ShortNodeString(child));*/
  }
} /* DoDFChildInPlace */

void z_bb3_WorkAndSplit(bb3_GlobalP bb3_global,
			bb3_WorkUnitP workUnit,
			s_seqP workList)
{
  z_bb_ProcSetP app_ps;

  /*printf("[%2d]WorkAndSplit\n", z_myproc); fflush(stdout);*/
  /*printf("[%2d]Nodes: %d %c\n", z_myproc, z_numNodes[workUnit->tag]
	 - s_seq_numElements(workList), glob.solNode?'+':'-');*/
  glob.psp = app_ps = (z_bb_ProcSetP) z_GetProcSet(bb3_global->bb3_psn);
  glob.bb3_glob = bb3_global;
  glob.workList = workList;
  if (app_ps->bb_BranchInPlace)
    app_ps->bb_BranchInPlace(bb3_global->bb3_app_global,
			     workUnit, glob.solNode,
			     (z_bb_DoProcP) DoDFChildInPlace, &glob);
  else
    app_ps->bb_Branch(bb3_global->bb3_app_global,
		      workUnit, glob.solNode,
		      (z_bb_DoProcP) DoDFChild, &glob);
  z_FreeNode(workUnit);
} /* z_bb3_WorkAndSplit */


void z_bb3_Fini(void)
{
  /*printf("[%2d]bb3_fini\n", z_myproc);*/
} /* z_bb3_Fini */

#ifndef SEQUENTIAL
void z_bb3_InitJob(void)
     /* started as parallel job by z_BranchBoundDFExp in order to
	broadcast the initial upper bound */
{
  /*printf("[%2d]bb3_init\n", z_myproc);*/
  glob.solNode = z_GetNodeBuffer();
} /* z_bb3_InitJob */

void z_bb3_GetResultJob(void)
     /* started as parallel job by z_BranchBoundDFExp and
	z_RestartBranchBoundDFExp at the end of the computation
	to copy the result (final upper bound) to the front end */
{
  if (z_myproc == 0) z_AddNodeBuffer(glob.solNode);
  if (glob.solNode) z_FreeNode(glob.solNode);
} /* z_bb3_GetResultJob */
#endif /* SEQUENTIAL */

z_NodeP z_BranchBoundDFExp(z_ProcSetNum procsetnum,
			   z_bb_GlobalDataP global,
			   z_NodeP root, z_NodeP initialUpperBound,
			   boolean noisy,
			   boolean sequential)
{
  bb3_GlobalP bb3_global;
  
  assert(bb3_inited);
  assert(z_frontend);
#ifdef SEQUENTIAL
  glob.solNode = z_NewCopyNode(initialUpperBound);
#else
  z_AddNodeBuffer(initialUpperBound);
  z_StartParallelJob(initJob, NULL);
  z_WaitWork();
#endif /* SEQUENTIAL */
  bb3_global = NewBB3Global(procsetnum, global);
  z_lb_Work(bb3_procsetnum, bb3_global, root, sequential);
#ifdef SEQUENTIAL
  return glob.solNode;
#else
  z_StartParallelJob(getResultJob, NULL); /* get result from proc 0 */
  z_WaitWork();
  return z_GetNodeBuffer();
#endif /* SEQUENTIAL */
} /* z_BranchBoundDFExp */


z_NodeP z_RestartBranchBoundDFExp(z_ProcSetNum procsetnum,
				  z_bb_GlobalDataP global,
				  boolean sequential)
{
  bb3_GlobalP bb3_global;
  
  assert(bb3_inited);
  /* z_StartParallelJob(equalizeUpperBoundJob, NULL); */
  /*z_WaitWork();*/
  bb3_global = NewBB3Global(procsetnum, global);
  z_lb_RestartWork(bb3_procsetnum, bb3_global, sequential);
#ifdef SEQUENTIAL
  return glob.solNode;
#else
  z_StartParallelJob(getResultJob, NULL); /* get result from proc 0 */
  z_WaitWork();
  return z_GetNodeBuffer();
#endif /* SEQUENTIAL */
} /* z_RestartBranchBoundDFExp */


/* Initialisation ------------------------------------------------------- */

void z_bb3_Init(void)
{
  /* called on every processor by initproc() */

  if (! bb3_inited) {
    bb3_inited = 1;
    z_lb_Init();
    z_lb_InstallCheckpointed((z_NodeP*)&glob.solNode);
    bb3_GlobalTag = z_InstallClass(bb3_Global, 1, 0, 0);
    bb3_procset.lb_Work = NULL;
    bb3_procset.lb_WorkAndSplit =  (z_lb_SplitProc) z_bb3_WorkAndSplit;
    bb3_procset.lb_Split = NULL;
    bb3_procset.lb_Fini = z_bb3_Fini;
    bb3_procsetnum = z_InstallProcSet((z_ProcSetP) &bb3_procset);
#ifndef SEQUENTIAL
    initJob = z_InstallJob(z_bb3_InitJob);
    getResultJob = z_InstallJob(z_bb3_GetResultJob);
    newUBTag = z_InstallMessage(NewUBHandler);
#endif /* SEQUENTIAL */
    glob.ub_changed = 0;
  }
} /* z_bb3_Init */

/* To do:

   equalize upper bounds at restart
   */
